<?php

/**
 * Copyright (C) 2014-2025 Textalk and contributors.
 * This file is part of Websocket PHP and is free software under the ISC License.
 */

namespace WebSocket;

/**
 * WebSocket\Connection class.
 * A client/server connection, wrapping socket stream.
 */
interface Constant
{
    public const GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
}
