<?php

// autoload_psr4.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'WebSocket\\' => array($vendorDir . '/phrity/websocket/src'),
    'Psr\\Log\\' => array($vendorDir . '/psr/log/src'),
    'Psr\\Http\\Message\\' => array($vendorDir . '/psr/http-factory/src', $vendorDir . '/psr/http-message/src'),
    'Phrity\\Util\\' => array($vendorDir . '/phrity/util-errorhandler/src'),
    'Phrity\\Net\\' => array($vendorDir . '/phrity/net-uri/src', $vendorDir . '/phrity/net-stream/src'),
    'Phrity\\Http\\' => array($vendorDir . '/phrity/http/src'),
    'Phrity\\Comparison\\' => array($vendorDir . '/phrity/comparison/src'),
    'Karriere\\PdfMerge\\' => array($vendorDir . '/karriere/pdf-merge/src'),
);
