<?php

/**
 * Copyright (C) 2014-2025 Textalk and contributors.
 * This file is part of Websocket PHP and is free software under the ISC License.
 */

namespace WebSocket\Exception;

use RuntimeException;

/**
 * WebSocket\Exception\Exception abstract class.
 * Core exception for repo
 */
abstract class Exception extends RuntimeException implements ExceptionInterface
{
}
