<?php

/**
 * File for IncomparableException.
 * @package Phrity > Comparison
 */

namespace Phrity\Comparison;

use InvalidArgumentException;

/**
 * Exception that should be thrown when instances can not be compared.
 */
class IncomparableException extends InvalidArgumentException
{
}
