<?php

/**
 * Copyright (C) 2014-2025 Textalk and contributors.
 * This file is part of Websocket PHP and is free software under the ISC License.
 */

namespace WebSocket\Exception;

/**
 * WebSocket\Exception\ConnectionLevelInterface interface.
 * Indicates error on connection level - connection should be closed
 */
interface ConnectionLevelInterface extends ExceptionInterface
{
}
