<?php return array(
    'root' => array(
        'name' => '__root__',
        'pretty_version' => 'dev-main',
        'version' => 'dev-main',
        'reference' => '9ce3413e78d2809619557cbe6ea79fafc5dd70c5',
        'type' => 'library',
        'install_path' => __DIR__ . '/../../',
        'aliases' => array(),
        'dev' => true,
    ),
    'versions' => array(
        '__root__' => array(
            'pretty_version' => 'dev-main',
            'version' => 'dev-main',
            'reference' => '9ce3413e78d2809619557cbe6ea79fafc5dd70c5',
            'type' => 'library',
            'install_path' => __DIR__ . '/../../',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'karriere/pdf-merge' => array(
            'pretty_version' => 'dev-master',
            'version' => 'dev-master',
            'reference' => '2903a4fda88466527bbafdfcc4d044e2b00a872f',
            'type' => 'library',
            'install_path' => __DIR__ . '/../karriere/pdf-merge',
            'aliases' => array(
                0 => '9999999-dev',
            ),
            'dev_requirement' => false,
        ),
        'phrity/comparison' => array(
            'pretty_version' => '1.4.1',
            'version' => '1.4.1.0',
            'reference' => 'cf80abb822537eeaaeb4142157cd667ca6372a13',
            'type' => 'library',
            'install_path' => __DIR__ . '/../phrity/comparison',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'phrity/http' => array(
            'pretty_version' => '1.0.1',
            'version' => '1.0.1.0',
            'reference' => '8f9d8eb92cd34cf7b9ec092b45730ec3ba668937',
            'type' => 'library',
            'install_path' => __DIR__ . '/../phrity/http',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'phrity/net-stream' => array(
            'pretty_version' => '2.3.2',
            'version' => '2.3.2.0',
            'reference' => '6d220446043c0b2bf512cfd89caa2494ad9afdd1',
            'type' => 'library',
            'install_path' => __DIR__ . '/../phrity/net-stream',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'phrity/net-uri' => array(
            'pretty_version' => '2.2.1',
            'version' => '2.2.1.0',
            'reference' => '0737de026b75177ae302ac9fdbbd0ffc2610f3b8',
            'type' => 'library',
            'install_path' => __DIR__ . '/../phrity/net-uri',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'phrity/util-errorhandler' => array(
            'pretty_version' => '1.2.2',
            'version' => '1.2.2.0',
            'reference' => '70a669cc22db2eed6a109ec66fd95168a4332c9b',
            'type' => 'library',
            'install_path' => __DIR__ . '/../phrity/util-errorhandler',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'phrity/websocket' => array(
            'pretty_version' => '3.6.2',
            'version' => '3.6.2.0',
            'reference' => 'b9816ed2b4a10c8c42bd0b6398044ab506869756',
            'type' => 'library',
            'install_path' => __DIR__ . '/../phrity/websocket',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'psr/http-factory' => array(
            'pretty_version' => '1.1.0',
            'version' => '1.1.0.0',
            'reference' => '2b4765fddfe3b508ac62f829e852b1501d3f6e8a',
            'type' => 'library',
            'install_path' => __DIR__ . '/../psr/http-factory',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'psr/http-message' => array(
            'pretty_version' => '2.0',
            'version' => '2.0.0.0',
            'reference' => '402d35bcb92c70c026d1a6a9883f06b2ead23d71',
            'type' => 'library',
            'install_path' => __DIR__ . '/../psr/http-message',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'psr/log' => array(
            'pretty_version' => '3.0.2',
            'version' => '3.0.2.0',
            'reference' => 'f16e1d5863e37f8d8c2a01719f5b34baa2b714d3',
            'type' => 'library',
            'install_path' => __DIR__ . '/../psr/log',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'tecnickcom/tcpdf' => array(
            'pretty_version' => '6.10.0',
            'version' => '6.10.0.0',
            'reference' => 'ca5b6de294512145db96bcbc94e61696599c391d',
            'type' => 'library',
            'install_path' => __DIR__ . '/../tecnickcom/tcpdf',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
    ),
);
